@@ -1,8 +1,8 @@
-use crate::linalg::{lu_pl, Lapack};
+use crate::linalg::{self, qr, svd, svddc, Lapack, LayoutError};
 use crate::DecompositionError;
 use itertools::izip;
-use ndarray::{s, Array1, Array2, ArrayBase, Axis, Data, Ix2, OwnedRepr, ScalarOperand};
-use ndarray_linalg::{error::LinalgError, QRInto, Scalar, UVTFlag, SVD, SVDDC};
+use lair::{decomposition::lu, Scalar};
+use ndarray::{s, Array1, Array2, ArrayBase, AssignElem, Axis, Data, Ix2, ScalarOperand};
 use num_traits::{real::Real, FromPrimitive};
 use rand::{Rng, RngCore, SeedableRng};
 use rand_distr::StandardNormal;
@@ -68,7 +68,7 @@ where
 
 impl<A> Pca<A>
 where
-    A: Scalar + Lapack,
+    A: FromPrimitive + Lapack,
     A::Real: ScalarOperand,
 {
     /// Returns the principal axes in feature space.
@@ -106,9 +106,10 @@ where
     ///
     /// # Errors
     ///
-    /// * `DecompositionError::InvalidInput` if any of the dimensions of `input`
-    ///   is less than the number of components.
-    /// * `DecompositionError::LinalgError` if the underlying Singular Vector
+    /// * [`DecompositionError::InvalidInput`] if any of the dimensions of
+    ///   `input` is less than the number of components, or the layout of
+    ///   `input` is incompatible with LAPACK.
+    /// * [`DecompositionError::LinalgError`] if the underlying Singular Vector
     ///   Decomposition routine fails.
     pub fn fit<S>(&mut self, input: &ArrayBase<S, Ix2>) -> Result<(), DecompositionError>
     where
@@ -142,8 +143,11 @@ where
     ///
     /// # Errors
     ///
-    /// Returns `DecompositionError::LinalgError` if the underlying Singular
-    /// Vector Decomposition routine fails.
+    /// * [`DecompositionError::InvalidInput`] if any of the dimensions of
+    ///   `input` is less than the number of components, or the layout of
+    ///   `input` is incompatible with LAPACK.
+    /// * [`DecompositionError::LinalgError`] if the underlying Singular Vector
+    ///   Decomposition routine fails.
     pub fn fit_transform<S>(
         &mut self,
         input: &ArrayBase<S, Ix2>,
@@ -164,7 +168,7 @@ where
     ///
     /// # Errors
     ///
-    /// Returns `DecompositionError::InvalidInput` if the number of rows of
+    /// Returns [`DecompositionError::InvalidInput`] if the number of rows of
     /// `input` is different from that of the training data, or the number of
     /// columns of `input` is different from the number of components.
     pub fn inverse_transform<S>(
@@ -181,16 +185,20 @@ where
     ///
     /// # Errors
     ///
-    /// * `DecompositionError::InvalidInput` if any of the dimensions of `input`
-    ///   is less than the number of components.
-    /// * `DecompositionError::LinalgError` if the underlying Singular Vector
+    /// * [`DecompositionError::InvalidInput`] if any of the dimensions of
+    ///   `input` is less than the number of components, or the layout of
+    ///   `input` is incompatible with LAPACK.
+    /// * [`DecompositionError::LinalgError`] if the underlying Singular Vector
     ///   Decomposition routine fails.
     fn inner_fit<S>(&mut self, input: &ArrayBase<S, Ix2>) -> Result<Array2<A>, DecompositionError>
     where
         S: Data<Elem = A>,
     {
         if input.shape().iter().any(|v| *v < self.n_components()) {
-            return Err(DecompositionError::InvalidInput);
+            return Err(DecompositionError::InvalidInput(format!(
+                "every dimension should be at least {}",
+                self.n_components()
+            )));
         }
 
         let means = if self.centering {
@@ -203,13 +211,12 @@ where
             Array1::zeros(input.ncols())
         };
 
-        let (u, sigma, vt) = if self.centering {
-            (input - &means).svd(true, true)?
+        let (mut u, sigma, vt) = if self.centering {
+            svd(&mut (input - &means), true)?
         } else {
-            input.svd(true, true)?
+            svd(&mut input.to_owned(), true)?
         };
 
-        let mut u = u.expect("`svd` should return `u`");
         let mut vt = vt.expect("`svd` should return `vt`");
         svd_flip(&mut u, &mut vt);
         self.total_variance = sigma.dot(&sigma);
@@ -374,8 +381,8 @@ where
 
 impl<A, R> RandomizedPca<A, R>
 where
-    A: Scalar + Lapack,
-    A::Real: ScalarOperand,
+    A: Scalar + FromPrimitive + Lapack,
+    A::Real: ScalarOperand + FromPrimitive,
     R: Rng,
 {
     /// Returns the principal axes in feature space.
@@ -413,9 +420,10 @@ where
     ///
     /// # Errors
     ///
-    /// * `DecompositionError::InvalidInput` if any of the dimensions of `input`
-    ///   is less than the number of components.
-    /// * `DecompositionError::LinalgError` if the underlying Singular Vector
+    /// * [`DecompositionError::InvalidInput`] if any of the dimensions of
+    ///   `input` is less than the number of components, or the layout of
+    ///   `input` is incompatible with LAPACK.
+    /// * [`DecompositionError::LinalgError`] if the underlying Singular Vector
     ///   Decomposition routine fails.
     pub fn fit<S>(&mut self, input: &ArrayBase<S, Ix2>) -> Result<(), DecompositionError>
     where
@@ -429,7 +437,7 @@ where
     ///
     /// # Errors
     ///
-    /// * `DecompositionError::InvalidInput` if the number of features in
+    /// * [`DecompositionError::InvalidInput`] if the number of features in
     ///   `input` does not match that of the training data.
     pub fn transform<S>(&self, input: &ArrayBase<S, Ix2>) -> Result<Array2<A>, DecompositionError>
     where
@@ -449,8 +457,11 @@ where
     ///
     /// # Errors
     ///
-    /// Returns `DecompositionError::LinalgError` if the underlying Singular
-    /// Vector Decomposition routine fails.
+    /// * [`DecompositionError::InvalidInput`] if any of the dimensions of
+    ///   `input` is less than the number of components, or the layout of
+    ///   `input` is incompatible with LAPACK.
+    /// * [`DecompositionError::LinalgError`] if the underlying Singular Vector
+    ///   Decomposition routine fails.
     pub fn fit_transform<S>(
         &mut self,
         input: &ArrayBase<S, Ix2>,
@@ -471,7 +482,7 @@ where
     ///
     /// # Errors
     ///
-    /// Returns `DecompositionError::InvalidInput` if the number of rows of
+    /// Returns [`DecompositionError::InvalidInput`] if the number of rows of
     /// `input` is different from that of the training data, or the number of
     /// columns of `input` is different from the number of components.
     pub fn inverse_transform<S>(
@@ -488,16 +499,20 @@ where
     ///
     /// # Errors
     ///
-    /// * `DecompositionError::InvalidInput` if any of the dimensions of `input`
-    ///   is less than the number of components.
-    /// * `DecompositionError::LinalgError` if the underlying Singular Vector
+    /// * [`DecompositionError::InvalidInput`] if any of the dimensions of
+    ///   `input` is less than the number of components, or the layout of
+    ///   `input` is incompatible with LAPACK.
+    /// * [`DecompositionError::LinalgError`] if the underlying Singular Vector
     ///   Decomposition routine fails.
     fn inner_fit<S>(&mut self, input: &ArrayBase<S, Ix2>) -> Result<Array2<A>, DecompositionError>
     where
         S: Data<Elem = A>,
     {
         if input.shape().iter().any(|v| *v < self.n_components()) {
-            return Err(DecompositionError::InvalidInput);
+            return Err(DecompositionError::InvalidInput(format!(
+                "every dimension should be at least {}",
+                self.n_components()
+            )));
         }
 
         let means = if self.centering {
@@ -652,18 +667,17 @@ fn randomized_svd<A, S, R>(
     input: &ArrayBase<S, Ix2>,
     n_components: usize,
     rng: &mut R,
-) -> Result<Svd<A>, LinalgError>
+) -> Result<Svd<A>, linalg::Error>
 where
     A: Scalar + Lapack,
+    A::Real: FromPrimitive,
     S: Data<Elem = A>,
     R: RngCore,
 {
     let n_random = n_components + 10; // oversample by 10
     let q = randomized_range_finder(input, n_random, 7, rng)?;
-    let b = q.t().dot(input);
-    let (u, sigma, vt) = b.svddc(UVTFlag::Some)?;
-    let u = u.expect("`svddc` should return `u`");
-    let mut vt = vt.expect("`svddc` should return `vt`");
+    let mut b = q.t().dot(input);
+    let (u, sigma, mut vt) = svddc(&mut b)?;
     let mut u = q.dot(&u);
     svd_flip(&mut u, &mut vt);
     Ok((u, sigma, vt))
@@ -675,28 +689,29 @@ fn randomized_range_finder<A, S, R>(
     size: usize,
     n_iter: usize,
     rng: &mut R,
-) -> Result<Array2<A>, LinalgError>
+) -> Result<Array2<A>, LayoutError>
 where
     A: Scalar + Lapack,
+    A::Real: FromPrimitive,
     S: Data<Elem = A>,
     R: RngCore,
 {
     let mut q = ArrayBase::from_shape_fn((input.ncols(), size), |_| {
         let r = A::Real::from_f64(rng.sample(StandardNormal))
             .expect("float to float conversion never fails");
-        A::from_real(r)
+        r.into()
     });
     let mut pl = q.view();
     q = input.dot(&pl);
     for _ in 0..n_iter {
-        lu_pl(&mut q)?;
+        q = lu::Factorized::from(q).into_pl();
         pl = q.slice(s![.., 0..cmp::min(q.nrows(), q.ncols())]);
         q = input.t().dot(&pl);
-        lu_pl(&mut q)?;
+        q = lu::Factorized::from(q).into_pl();
         pl = q.slice(s![.., 0..cmp::min(q.nrows(), q.ncols())]);
         q = input.dot(&pl);
     }
-    let (q, _) = q.qr_into()?;
+    let q = qr(q)?;
     Ok(q)
 }
 
@@ -704,8 +719,8 @@ where
 ///
 /// # Errors
 ///
-/// * `DecompositionError::InvalidInput` if the number of features in
-///   `input` does not match that of the training data.
+/// * [`DecompositionError::InvalidInput`] if the number of features in `input`
+///   does not match that of the training data.
 fn transform<A, S>(
     input: &ArrayBase<S, Ix2>,
     components: &Array2<A>,
@@ -717,7 +732,10 @@ where
     S: Data<Elem = A>,
 {
     if input.ncols() != means.len() {
-        return Err(DecompositionError::InvalidInput);
+        return Err(DecompositionError::InvalidInput(format!(
+            "# of columns should be {}",
+            means.len()
+        )));
     }
 
     let transformed = if centering {
@@ -733,7 +751,7 @@ where
 ///
 /// # Errors
 ///
-/// Returns `DecompositionError::LinalgError` if the underlying Singular
+/// Returns [`DecompositionError::LinalgError`] if the underlying Singular
 /// Vector Decomposition routine fails.
 fn transform_with_u<A, S>(
     u: &Array2<A>,
@@ -745,25 +763,24 @@ where
     A: Scalar,
     S: Data<Elem = A>,
 {
-    let mut y: ArrayBase<OwnedRepr<A>, Ix2> = // initialized in the following `for` loop
-        unsafe { ArrayBase::uninitialized((input.nrows(), n_components)) };
+    let mut y = Array2::<A>::uninit((input.nrows(), n_components));
     for (y_row, u_row) in y
         .lanes_mut(Axis(1))
         .into_iter()
         .zip(u.slice(s![.., 0..n_components]).lanes(Axis(1)))
     {
         for (y_v, u_v, sigma_v) in izip!(y_row.into_iter(), u_row, singular) {
-            *y_v = *u_v * A::from_real(*sigma_v);
+            y_v.assign_elem(*u_v * (*sigma_v).into());
         }
     }
-    y
+    unsafe { y.assume_init() }
 }
 
 /// Transforms data back to its original space.
 ///
 /// # Errors
 ///
-/// Returns `DecompositionError::InvalidInput` if the number of rows of
+/// Returns [`DecompositionError::InvalidInput`] if the number of rows of
 /// `input` is different from that of the training data, or the number of
 /// columns of `input` is different from the number of components.
 fn inverse_transform<A, S>(
@@ -777,7 +794,10 @@ where
     S: Data<Elem = A>,
 {
     if input.ncols() != components.nrows() {
-        return Err(DecompositionError::InvalidInput);
+        return Err(DecompositionError::InvalidInput(format!(
+            "# of columns should be {}",
+            components.nrows()
+        )));
     }
 
     let inverse_transformed = if centering {
@@ -816,7 +836,7 @@ where
             };
         }
         if signum < A::zero().re() {
-            let signum = A::from_real(signum);
+            let signum = signum.into();
             for e in u_col {
                 *e *= signum;
             }
@@ -829,7 +849,7 @@ where
 
 #[cfg(test)]
 mod test {
-    use approx::{assert_abs_diff_eq, assert_relative_eq, AbsDiffEq};
+    use approx::{assert_abs_diff_eq, assert_relative_eq};
     use ndarray::{arr2, Array2};
     use rand::Rng;
     use rand_distr::StandardNormal;
@@ -913,7 +933,6 @@ mod test {
     #[test]
     #[cfg(feature = "serde")]
     fn pca_serialize() {
-        use approx::AbsDiffEq;
         let mut pca = super::Pca::new(1);
         let x = arr2(&[[1_f32, 1_f32]]);
         assert!(pca.fit(&x).is_ok());
@@ -1008,7 +1027,6 @@ mod test {
     #[test]
     #[cfg(feature = "serde")]
     fn randomized_pca_serialize() {
-        use approx::AbsDiffEq;
         let mut pca = super::RandomizedPca::with_seed(1, RNG_SEED);
         let x = arr2(&[[1_f32, 1_f32]]);
         assert!(pca.fit(&x).is_ok());
